%{


  /***************************************************************************

Copyright 2016 J.-L. LAMOTTE

This file is part of CADTRACE.

    CADTRACE is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    CADTRACE is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with CADTRACE. If not, see
    <http:#www.gnu.org/licenses/>.


**************************************************************************/


#include <stdlib.h>
#include <string.h>

#include "yacc_cadtrace.h"

  #define DEBUGn
  #define DEBUG1n
  #define DEBUG2n

static  int nbparanthese=0;

//  #define DEBUT_LIGNE 1
//  #define NOM_FONCTION 2
//  #define NOM_FICHIER 3
//  #define NUMERO_LIGNE 4
//  #define FDF 5


%}


%x DEBLIG
%x APRESNOMFONC
%x ARGFONCTION
%x APRESARG




%%





^((#[0-9]*)|(">0"))" "* {
#ifdef DEBUG
    printf("LEX : APPEL chaine %s  entier  #%d\n",yytext,atoi(yytext+1)); 
#endif
    yylval.entier=atoi(yytext+1);
    BEGIN(DEBLIG);
    return(DEBUT_LIGNE);
}

^("(gdb)  > > >#0")" "* {
#ifdef DEBUG
  printf("Pour gérer la première ligne dans certains cas : (gdb) > > >#0");
  printf("LEX : APPEL chaine %s  entier  #%d\n",yytext,atoi(yytext+1)); 
#endif
    yylval.entier=0;;
    BEGIN(DEBLIG);
    return(DEBUT_LIGNE);
}

^(("=>"\[[0-9]*\])|("  "\[[0-9]*\]))" "* {
#ifdef DEBUG
    printf("LEX : APPEL [%d]\n",atoi(yytext+3)-1); 
#endif
    yylval.entier=atoi(yytext+1);
    BEGIN(DEBLIG);
    return(DEBUT_LIGNE);
}

<DEBLIG>" "* {
#ifdef DEBUG2
  printf(">%s<\n",yytext);
#endif
}

<DEBLIG>0x[0-9a-f]+ {
/* on ne retourne pas les adresses */
#ifdef DEBUG
    printf("LEX :  adresse %s\n",yytext); 
#endif
}

<DEBLIG>" "*"in"" "* {
/* on ne retourne pas les in */
#ifdef DEBUG
    printf("LEX :  in\n");
#endif
}


<DEBLIG>[a-zA-Z_][a-zA-Z_0-9.:]*[\+\-\*\/\=\<\>]* {
#ifdef DEBUG
    printf("LEX : NOM FONCTION %s\n",yytext);
#endif
    yylval.chaine = (char*)strdup(yytext);
    if (strncmp( yylval.chaine, "_start",6)==0) {
      BEGIN(INITIAL);
      //printf("INITIAL\n");
    }
    else   BEGIN(APRESNOMFONC);
    return(NOM_FONCTION);
}



<APRESNOMFONC>("("|" (") { 
#ifdef DEBUG1
    printf("LEX : PARANTHESE OUVRANTE [%s]\n",yytext);
#endif
    nbparanthese++;
    int c;
    while ((c=input())!=')');

    c=input();
    while (c==' '|| c=='\n')
          c=input();
    unput(c);

    BEGIN(APRESARG);
}

<APRESARG>((#[0-9]*)|(">0"))" \n"*  {
  yylval.entier=atoi(yytext+1);
#ifdef DEBUG2
  printf("TROUVER #XXXXX apres nom de fichier. ON repart \n");
#endif
  BEGIN(DEBLIG);
  return(DEBUT_LIGNE);
 }

<APRESARG>[\n" "]*"Breakpoint" { 
#ifdef DEBUG1
  printf("LEX : line : [%s]\n",yytext);
#endif
  BEGIN(INITIAL);
  
}


<APRESARG>([\n" "]*"at"" "*|" "*"at"" "*0x[0-9a-f]+" "*) {
#ifdef DEBUG1
  printf("LEX : at : [%s]\n",yytext);
#endif
  
}

<APRESARG>" "*"line"" "* {
#ifdef DEBUG1
  printf("LEX : line : [%s]\n",yytext);
#endif
  
}

<APRESARG>" "*"from"" "* {
#ifdef DEBUG1
  printf("LEX : line : [%s]\n",yytext);
#endif
  
}

<APRESARG>":" {
#ifdef DEBUG1
printf("LEX : :\n",yytext);
#endif
}

<APRESARG>" "*"in"" "* {
#ifdef DEBUG1
printf("LEX : in \n",yytext);
#endif
}

<APRESARG>\"*[a-zA-Z/+\-*~#$%.][a-zA-Z_0-9./+\-*~#$%]*\"* { 
#ifdef DEBUG
  printf("LEX : NOM FICHIER <%s>\n",yytext);
#endif
  yylval.chaine = (char*)strdup(yytext);
  return(NOM_FICHIER);
}


<APRESARG>" "*  {
  // SERT A RIEN 
#ifdef DEBUG2
  printf("<APRES ARG >%s<\n", yytext);
#endif
}



<APRESARG>\n {
#ifdef DEBUG2
  printf("passage a la ligne APRES ARG\n");
#endif
BEGIN(INITIAL);
}



<APRESARG>[0-9]+ {
#ifdef DEBUG
   printf("LEX : NUMERO DE LIGNE %d\n",atoi(yytext));  
#endif
   yylval.entier=atoi(yytext);
   BEGIN(INITIAL);
   return(NUMERO_LIGNE);
}



" "* {
#ifdef DEBUG2
  printf(">%s<\n",yytext);
#endif
}



\n {
#ifdef DEBUG2
  printf("passage a la ligne\n");
#endif
}

. {  
#ifdef DEBUG2
  printf("%c", (*yytext>=97 && *yytext<=122)?*yytext-32:*yytext); 
#endif

}


%%

// int main()
// {
//   for(;;){
//     int v=yylex();
//     if (v==FDF) exit(1);
//     printf("%d\n",v);
//   }
// }

int yywrap() 
{ 
  // pour sortir du parseur sur fin de fichier
  return(1);
}


