%{

 /***************************************************************************

Copyright 2016 J.-L. LAMOTTE

This file is part of CADTRACE.

    CADTRACE is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    CADTRACE is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with CADTRACE. If not, see
    <http:#www.gnu.org/licenses/>.


**************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "desc_compilateur.h"
  
char mul_error[]   = " UNSTABLE MULTIPLICATION.";
char sign_error[]  = " INSTABILITY IN SIGN FUNCTION.";
char mod_error[]   = " INSTABILITY IN MOD FUNCTION.";
char dim_error[]   = " INSTABILITY IN DIM FUNCTION.";
char int_error[]   = " INSTABILITY IN INT FUNCTION.";
char nint_error[]  = " INSTABILITY IN NINT FUNCTION.";
char div_error[]   = " UNSTABLE DIVISION.";
char test_error[]  = " UNSTABLE BRANCHING.";
char pui_error[]   = " UNSTABLE POWER FUNCTION.";
char log_error[]   = " NEGATIVE OR NULL VALUE IN LOG FUNCTION.";
char sqrt_error[]  = " NEGATIVE OR NULL VALUE IN SQRT FUNCTION.";
char exp_error[]   = " UNSTABLE EXPONENTIAL FUNCTION.";
char log10_error[] = " NEGATIVE OR NULL VALUE IN LOG10  FUNCTION.";
char perte_error[] = " LOSS OF ACCURACY DUE TO CANCELLATION(S).";
char abs_error[]   = " INSTABILITY IN ABS FUNCTION.";
  
 extern  void yyerror(char *);


extern int yylex(void);
extern char yytext[];

extern FILE *yyin;
extern int yyparse();
extern char *fonction_connue(char*);

int NbInstabilite=0;
int YaccErreur=0;


#define MAXFONCTION 1000
#define MAXFICHIER  1000
#define MAXAPPEL 25 
#define DECALAGE 2 /* pour l'affichage */

#define DEBUGn

/* tableau des nom de fonctions rencontrees */

char *nomfonction[MAXFONCTION];
int nbfonction=0;


/* tableau des nom de fichiers rencontrees */

char *nomfichier[MAXFICHIER];
int nbfichier=0;

/* Structure pour le tableau des appels */

struct monappel {
  /* caracteristique de l'appel */
  char *nfonc;
  char *nfic;
  int numligne;
  int nbappel;

  struct monappel **tabappel;
    int maxtabappel;      /* pour realloc quand le tableau est plein */
  int numtabappel;
}; 

typedef struct monappel Appel;


Appel *courant=NULL;
Appel debut={NULL,NULL,0,-1,NULL,0,0};

void bilan_appel();
void usage(char *);
char *insertion_tab_fonction(char *);
char *insertion_tab_fichier(char *);
 void recherche_insertion_appel(char *, char *, int);

 %}



%token  FDF

%union {
	int entier;
	char *chaine;
	void *li;
}	

%token <chaine>  NOM_FONCTION  NOM_FICHIER 
%token <entier> NUMERO_LIGNE DEBUT_LIGNE


%start begin 

%%

begin :  begin lignes
        | lignes
 	;



lignes : DEBUT_LIGNE NOM_FONCTION 
{
  char *fonctrouve=NULL;


#ifdef DEBUG
  printf("YACC : DEBUT LIGNE NOM_FONCTION  %d %s \n",$1, $2);
#endif

  /* on n'insert ps les appels  la fonction unstabiliy */
  if ($1!=0){       
     /* on recherche si le nom de la fonction est deja dans 
	le tableau des fonctions */

      if ($1==1){ 
      
	fonctrouve=fonction_connue($2);      
#ifdef DEBUG
	printf("YACC : NOM_FONCTION RECONNUE  <%s> :: %s \n", $2,fonctrouve);
#endif
	fonctrouve=insertion_tab_fonction(fonctrouve);	
	courant = &debut;
	/* on ne met pas le nom du fichier */
	recherche_insertion_appel(fonctrouve, NULL,-1);
	NbInstabilite++;
      }
      
      else {
	fonctrouve=insertion_tab_fonction($2);
	recherche_insertion_appel(fonctrouve, NULL, -1);
      }
#ifdef DEBUG
      printf("@@\n11\n");
#endif
  }
  free($2);
}
|
DEBUT_LIGNE NOM_FONCTION  NOM_FICHIER   
{

  char *fonctrouve=NULL;
  char *fichtrouve=NULL;
  
#ifdef DEBUG
  printf("YACC : DEBUT_LIGNE NOM_FONCTION NOM_FICHIER  %d %s  fichier  %d\n",
	 $1, $2, $3);
#endif
  
  /* on n'insert pas les appels aux fonctions de CADNA */
  if ($1!=0) {
    
    /* on recherche si le nom de la fonction est deja dans 
       le tableau des fonctions */
    fonctrouve=insertion_tab_fonction($2);
    
    /* on recherche si le nom de la fonction est deja dans 
       le tableau des fonctions */
    
    
    if ($1==1){ 

      fonctrouve=fonction_connue(fonctrouve);      
      
      courant = &debut;
      /* on ne met pas le nom du fichier */
      recherche_insertion_appel(fonctrouve, NULL,-1);
      NbInstabilite++;
    }
      else {
	fichtrouve=insertion_tab_fichier($3);
	recherche_insertion_appel(fonctrouve, fichtrouve,-1); 
      }
#ifdef DEBUG
    printf("@@\n22\n");
#endif
  }

  free($2);
  free($3);
}
|
DEBUT_LIGNE NOM_FONCTION  NOM_FICHIER NUMERO_LIGNE  
{

  char *fonctrouve=NULL;
  char *fichtrouve=NULL;
  
#ifdef DEBUG
  printf("YACC : DEBUT_LIGNE NOM_FONCTION NOM_FICHIER NUMERO_LIGNE %d %s  fichier %s ligne %d\n",
	 $1, $2, $3, $4);
#endif
  
  /* on n'insert pas les appels aux fonctions de CADNA */
  if ($1!=0) {
    
    /* on recherche si le nom de la fonction est deja dans 
       le tableau des fonctions */
    fonctrouve=insertion_tab_fonction($2);
    
    /* on recherche si le nom de la fonction est deja dans 
       le tableau des fonctions */
    
    
    if ($1==1){ 

      fonctrouve=fonction_connue(fonctrouve);      
      
      courant = &debut;
      /* on ne met pas le nom du fichier */
      recherche_insertion_appel(fonctrouve, NULL,-1);
      NbInstabilite++;
    }
      else {
	fichtrouve=insertion_tab_fichier($3);
	recherche_insertion_appel(fonctrouve, fichtrouve,$4); 
      }
#ifdef DEBUG
    printf("@@\n33\n");
#endif
  }

  free($2);
  free($3);
}
| DEBUT_LIGNE NOM_FONCTION  NUMERO_LIGNE NOM_FICHIER  
{

  char *fonctrouve=NULL;
  char *fichtrouve=NULL;
  
#ifdef DEBUG
  printf("YACC : DEBUT_LIGNE NOM_FONCTION NUMERO_LIGNE NOM_FICHIER  %d %s  ligne %d fichier %s \n",
	 $1, $2, $3, $4);
#endif
  
  /* on n'insert pas les appels aux fonctions de CADNA */
  if ($1!=0) {
    
    /* on recherche si le nom de la fonction est deja dans 
       le tableau des fonctions */
    fonctrouve=insertion_tab_fonction($2);
    
    /* on recherche si le nom de la fonction est deja dans 
       le tableau des fonctions */
    
    
    if ($1==1){ 

      fonctrouve=fonction_connue(fonctrouve);      
      
      courant = &debut;
      /* on ne met pas le nom du fichier */
      recherche_insertion_appel(fonctrouve, NULL,-1);
      NbInstabilite++;
    }
      else {
	fichtrouve=insertion_tab_fichier($4);
	recherche_insertion_appel(fonctrouve, fichtrouve,$3); 
      }
#ifdef DEBUG
    printf("@@\n44\n");
#endif
  }

  free($2);
  free($4);
}


;



%%

/********************************************************************
Principe de base : On stocke dans le tableau nomfonction tous les noms
de fonctions et dans nomfichier tous les noms de fichiers. Ceci permet 
d'eviter une explosion de la memoire pour le stockage des noms.

Ensuite on remplit l'arbre des appels. 


 ********************************************************************/
/*void analyse_arg(int, **char , **char, **char, **char, **char, *int); */



/************************************************************************

Analyse des arguments du programme

 ************************************************************************/
void usage(char *s)
{
    fprintf(stderr,"\n");
    fprintf(stderr,"usage: %s [Option] gdb.out\n",s);
    fprintf(stderr,"gdb.out is obtained by executing your program with gdb (see README file)\n");
    //    fprintf(stderr,"      : %s [Option1] [Option2] \n\n",s);

    fprintf(stderr,"Option: \n");
    fprintf(stderr,"-n number: gives the depth of the call function tree to print\n");
    fprintf(stderr,"           (default 1)\n\n");

    /*    fprintf(stderr,"Option2 \n");
    fprintf(stderr,"  \n");
    fprintf(stderr,"-a name \\\"argument\\\"  : specify  the application name and the arguments between  \\\"\n");
    fprintf(stderr,"-s FileName      : specify the filename used to save teh trace\n");*/
    fprintf(stderr,"  \n");
    exit(1);
}

void analyse_arg(int argc, char **argv, char** fichiertrace, char** appli, char** argappli,
		 char** sauvetrace, int *niv)

{
    int i;
    char opt, chaine1[256], chaine2[256];
    FILE *f;

    if (argc==1) usage(argv[0]);

    for (i = 1; i < argc; i++){

	if (argv[i][0] != '-') {
	    *fichiertrace=strdup(argv[i]);
	    continue;
	}
	if (strlen(argv[i])!=2) usage(argv[0]);
	
	opt = argv[i][1];
	
	if (!opt) usage(argv[0]);
	
	if (strchr("sna", opt)) {
	    switch (opt) {
		
	    case 'n':
		if (i+1<argc) *niv=atoi(argv[++i]);
		if (niv<0) {
		    fprintf(stderr,"Niveau d'affichage < 0\n");
		    usage(argv[0]);
		}
	      break;
	      
	    case 's':
		if (i+1<argc) *sauvetrace=strdup(argv[++i]);
		break;
		
	    case 'a':
		if (i+1<argc) *appli=strdup(argv[++i]);
		if (i+1<argc) {
		    if (argv[i+1][0]=='"')
			*argappli=strdup(argv[++i]);
		}
		break;
	    default : usage(argv[0]);
	    }
	}
    }
    
    if (*appli && *fichiertrace ){
	fprintf(stderr,"Erreur dans les arguments. Le programme ne peut pas analyser un fichier\n");
	fprintf(stderr,"de trace et un programme en meme temps\n\n");
	usage(argv[0]);
    }
    if (*sauvetrace && *fichiertrace) {
	fprintf(stderr,"Erreur dans les arguments. L'option -s est incompatible avec l'analyse d'un\n");
	fprintf(stderr,"fichier de trace \n");
	usage(argv[0]);
    }

    if (*fichiertrace) {
	if ((f=fopen(*fichiertrace,"r"))==NULL){
	    fprintf(stderr,"Impossible d'ouvrir le fichier %s\n",*fichiertrace);
	    exit(1);
	}

	// test du type de trace
       	// fscanf(f,"%s %s",chaine1, chaine2);
// 	if (strcmp(chaine1,"GNU") || strcmp(chaine2,"gdb")){
// 	    fprintf(stderr,"Le fichier <%s> n'est pas une trace obtenue par gdb.\n",
// 		    *fichiertrace);
// 	    exit(1);	    
// 	}
    }
}



/********************************************************************************

Construction de l'arbre dea instabilites										 
 ********************************************************************************/


char *insertion_tab_fonction(char *f)
{
  int i;
  char *fonctrouve=NULL;

  /* on recherche si l'appel de f est deja dans le tableau des fonctions */

#ifdef DEBUG
  printf("insertion_tab_fonction :recherche insertion de <%s>\n",f);
#endif

  for(i=0;i<nbfonction;i++)
    if (!strcmp(nomfonction[i],f)) {
      fonctrouve=nomfonction[i];
      break;
    }

#ifdef DEBUG
  printf("insertion_tab_fonction : pointeur sur la fonction trouvee <%s> %p\n",f,fonctrouve);
#endif

  if (fonctrouve==NULL ){
    /* la fonction n'est pas trouvee. on l'insert dans le tableau des fonctions */

#ifdef DEBUG
  printf("insertion_tab_fonction : Fonction non trouvee recherche insertion de <%s>\n",f);
#endif

    if (nbfonction<MAXFONCTION) {
      nomfonction[nbfonction]=strdup(f);
      fonctrouve=nomfonction[nbfonction++];
    }
    else {
      fprintf(stderr,"il y a trop de fonctions pour ce programme\n"); 
      fprintf(stderr,"Appelez la maintenance\n");
      exit(1);
    }
  }
  
#ifdef DEBUG
  printf("insertion_tab_fonction :  <%s> -> %p \n",f,fonctrouve);
#endif
  return(fonctrouve);
}


char *insertion_tab_fichier(char *f)
{
  int i;
  char *fichtrouve=NULL;

  /* on recherche si l'appel de f est deja dans le tableau des fichiers */
#ifdef DEBUG
  printf("recherche insertion de %s\n",f);
#endif


  for(i=0;i<nbfichier;i++)
    if (!strcmp(nomfichier[i],f)) {
      fichtrouve=nomfichier[i];
      break;
    }

  if (fichtrouve==NULL ){
    /* la fichier n'est pas trouvee. on l'insert dans le tableau des fichiers */
    if (nbfichier<MAXFICHIER) {
      nomfichier[nbfichier]=strdup(f);
      fichtrouve=nomfichier[nbfichier++];
    }
    else {
      fprintf(stderr,"il y a trop de fichiers pour ce programme\n"); 
      fprintf(stderr,"Appelez la maintenance\n");
      exit(1);
    }
  }
  return(fichtrouve);
}

void recherche_insertion_appel(char *fo, char *fi, int li)
{
  Appel *c=courant, *appeltrouve=NULL;
  int i;

#ifdef DEBUG
  printf("recherche_insertion_appel : recherche fonction,fichier, ligne dans les appels\n");
#endif

/* on recherche si cet appel est deja reference dans les tables */
  

  for(i=0;i<c->numtabappel;i++){
#ifdef DEBUG
      printf("recherche_insertion_appel : %p %p   ---- %p %p  ---- %d %d\n",
	     c->tabappel[i]->nfonc,fo,
	     c->tabappel[i]->nfic,fi, 
	     c->tabappel[i]->numligne,li);
#endif
      if (c->tabappel[i]->nfonc==fo && 
	  c->tabappel[i]->nfic==fi  && 
	  c->tabappel[i]->numligne==li){
	  appeltrouve=c->tabappel[i];
	  break;
      }
  }
  
  
  if (appeltrouve){
#ifdef DEBUG
      printf("recherche_insertion_appel : TROUVE on ajoute 1\n",appeltrouve);
#endif
    (appeltrouve->nbappel)++;
    courant=appeltrouve;
  }
  else {
#ifdef DEBUG
      printf("recherche_insertion_appel :  NON TROUVE, insertion  nouvel element \n");
#endif  
 
   /* la fonction n'a pas ete trouvee dans les appels, on l'insert */

    if (c->numtabappel>=c->maxtabappel) {
#ifdef DEBUG
      printf("APPEL ALLOCATION du tableau \n");
#endif  	
      if ((c->tabappel=realloc(c->tabappel,sizeof(Appel *)*(c->maxtabappel+MAXAPPEL)))==NULL){	
	fprintf(stderr,"il y a trop d'appels de fonctions de base pour ce programme\n"); 
	fprintf(stderr,"Appelez la maintenance\n");
	exit(1);
      }
      c->maxtabappel+=MAXAPPEL;
    }

    if ((c->tabappel[c->numtabappel]=malloc(sizeof(Appel)))==NULL){
      fprintf(stderr,"Probleme allocation memoire 2\n"); 
      fprintf(stderr,"Appelez la maintenance\n");
      exit(1);
    }

    courant=c->tabappel[c->numtabappel];
    c->numtabappel++;
    c=courant;
    (c->nbappel)++;
    
    c->nfonc=fo;
    c->nfic=fi;
    c->nbappel=1;
    c->numligne=li;
    c->tabappel=NULL;
    c->maxtabappel=0;
    c->numtabappel=0;
  }
}



void origine_appel(char *fo, char *fi, int numlig)
{
  char *fonctrouve=NULL;
  char *fichtrouve=NULL;
  Appel *appeltrouve=NULL;
  int i;
  
  /* on recherche si l'appel de fi est deja dans le tableau des fonctions */
  fonctrouve=insertion_tab_fonction(fo);


  /* on recherche si l'appel de fo est deja dans le tableau des fichiers */
  if (fi) fichtrouve=insertion_tab_fichier(fi);

  /* on recherche si le triplet est deja dans la table des appels */
  recherche_insertion_appel(fonctrouve, fichtrouve, numlig);

}


/********************************************************************************

Affichage de l'arbre des instabilites										 
 ********************************************************************************/


void affiche(Appel* tab[], int nbelem, int e, 
	     int taillefonc, int taillefic,int niveau)
{
  int i; 
  char ch1[256], ch2[256];
  char s[]="";

/* on sort de la recursion quand nbelem==0 */

  if (e>=niveau) return; 

  if (!e) sprintf(ch1,"%%%dsThere are  %%%dd  %%%ds ",e*DECALAGE,6,taillefonc);
  else sprintf(ch1,"%%%dsout of which  %%%dd call(s) by <%%%ds> ",e*DECALAGE,6,taillefonc);
  sprintf(ch2,"file \"%%%ds\" line %%5d",taillefic);


  for(i=0;i<nbelem;i++){

    printf(ch1,s,tab[i]->nbappel, tab[i]->nfonc);
    if (tab[i]->nfic)
      printf(ch2, tab[i]->nfic,tab[i]->numligne);
    printf("\n");

    if (tab[i]->numtabappel) 
	affiche(tab[i]->tabappel,tab[i]->numtabappel,e+1,
		taillefonc,taillefic,niveau);
  }
/* pour mettre une ligne blanche apres le main() */
 if (e+1 >=niveau || !tab[i-1]->numtabappel) 
      printf("\n");
}

void bilan_appel(int niveau)
{
  int i;
  int taillenomfichier=0, taillenomfonction=0;
  

  for(i=0;i<nbfonction;i++){
    int s=strlen(nomfonction[i]);
    if (taillenomfonction<s) s=taillenomfonction;
  }

  for(i=0;i<nbfichier;i++){
    int s=strlen(nomfichier[i]);
    if (taillenomfichier<s) s=taillenomfichier;
  }

  printf("\n\nCADNA SOFTWARE --- Numerical instabilities check-up.\n");
  printf("--------------------------------------------------------\n");
  printf("\nThere are %d numerical instabilities.\n\n\n",
	 NbInstabilite);

  


  if (debut.numtabappel)
    affiche(debut.tabappel, debut.numtabappel,0,
	    taillenomfonction, taillenomfichier,niveau);
  
  if (YaccErreur){
      fprintf(stderr,"\n\nThe trace file is not readable by the tool.\n");
      exit(1);
  }

  if (!debut.numtabappel){
      fprintf(stderr,"\n\nThe trace file is empty. \n");
      exit(1);
  }


}

/********************************************************************************



 ********************************************************************************/

 void yyerror(char *s)

{
    fprintf(stderr,"ANALYSIS ERROR \n");
    fflush(stdout);
    YaccErreur =1;
}

int main(int argc, char *argv[])
{
    char *fichiertrace=NULL;
    char *appli=NULL;
    char* argappli=NULL;
    char *sauvetrace=NULL;
    int niv=1,res;
    char tmp[256];
    int traceaeffacer=0;

    analyse_arg(argc, argv, &fichiertrace, &appli, &argappli, &sauvetrace, &niv);


    if(!fichiertrace) {
	char *pos_lib;
	FILE *f;
	char *nfich_unst, *nfich_gdb, *nfich_trace;
	char nfich_unstXX[]="cad_XXXXXX", nfich_gdbXX[]="cad_XXXXXX";
	char nomfonc[256];



	nfich_unst=mktemp(nfich_unstXX);
	nfich_gdb=mktemp(nfich_gdbXX);
	
	/* on fait le nm sur la librairie */

	if ((f=fopen(nfich_gdb,"w"))==NULL){
	    fprintf(stderr,"Impossible to analyze the CADNA trace 2\n");
	    fprintf(stderr,"Problem to create a file\n");
	    exit(1);
	}
	    
#ifdef SUN
	fprintf(f,"break unstability\n");
#else
	fprintf(f,"set pagination off\n");
	fprintf(f,"break instability\n");
	fprintf(f,"set $_exitcode = -999\n");
#endif
	if (argappli) fprintf(f,"run %s\n",argappli);
	else fprintf(f,"run\n");
	fprintf(f,"while $_exitcode==-999\nwhere\ncont\nend\nquit");
	
	fclose(f);
	
	/* si l'utilisateur ne veut pas de fichier trace */

	if (sauvetrace)
	    fichiertrace=sauvetrace;
	else {
	    fichiertrace=nfich_unst;
	    traceaeffacer=1;
	}

	/* execution de gdb */

	sprintf(tmp,"gdb   -x %s %s >%s ", nfich_gdb, appli,fichiertrace);
	
	printf("Launch the program with gdb\n");
        res=system(tmp);
	if (res==-1 || res==127) {
	    fprintf(stderr,"Problem during the run.\n");
	    exit(1);
	}	

       	sprintf(tmp,"rm -f  %s",nfich_gdb);
        res=system(tmp);

	printf("Analysis of the running trace\n");

    }

    if ((yyin=fopen(fichiertrace,"r"))==NULL){
	fprintf(stderr,"Erreur ouverture fichier %s   \n", fichiertrace);
	return 0;
    }

   yyparse();
  
   /* l'utilisateur ne veut pas conserver son fichier trace */

#ifdef DEBU
      printf("FIN DE FICHIER");
#endif


   if (traceaeffacer){
       sprintf(tmp,"rm -f  %s",fichiertrace);
       res=system(tmp);
   }
   

     bilan_appel(niv+1);
   if (YaccErreur) {
       printf("\n\nAn error occurs during the run. All the instabilities are not printed\n"); 
   } 
}                 


char *fonction_connue(char *fonctrouve)
{
  
  if (strstr(fonctrouve,FONC_ADD)) fonctrouve=perte_error;
  else if (strstr(fonctrouve,FONC_SUB)) fonctrouve=perte_error;
  else if (strstr(fonctrouve,FONC_MUL)) fonctrouve=mul_error;
  else if (strstr(fonctrouve,FONC_DIV))  fonctrouve=div_error;
  else if (strstr(fonctrouve,FONC_LE)) fonctrouve=test_error;
  else if (strstr(fonctrouve,FONC_LT)) fonctrouve=test_error;
  else if (strstr(fonctrouve,FONC_GT)) fonctrouve=test_error;
  else if (strstr(fonctrouve,FONC_GE)) fonctrouve=test_error;
  else if (strstr(fonctrouve,FONC_INT)) fonctrouve=int_error;
  else if (strstr(fonctrouve,FONC_NINT)) fonctrouve=nint_error;
  else if (strstr(fonctrouve,FONC_ABS)) fonctrouve=abs_error;
  else if (strstr(fonctrouve,FONC_MAX)) fonctrouve=test_error;
  else if (strstr(fonctrouve,FONC_MIN)) fonctrouve=test_error;
  else if (strstr(fonctrouve,FONC_SIGN)) fonctrouve=sign_error;
  else if (strstr(fonctrouve,FONC_DIM)) fonctrouve=dim_error;
  else if (strstr(fonctrouve,FONC_POW)) fonctrouve=pui_error;
  else if (strstr(fonctrouve,FONC_LOG)) fonctrouve=log_error;
  else if (strstr(fonctrouve,FONC_LOG10)) fonctrouve=log10_error;
  else if (strstr(fonctrouve,FONC_SQRT)) fonctrouve=sqrt_error;
  else if (strstr(fonctrouve,FONC_EXP)) fonctrouve=exp_error;

  
  return fonctrouve;
}



